/*  Felimage Noise Plugin for the GIMP
 *  Copyright (C) 2005 Guillermo Romero Franco <drirr_gato@users.sourceforge.net>
 *
 *  This file is part of the Felimage Noise Plugin for the GIMP
 *  
 *  Felimage Noise Plugin for the Gimp is free software; 
 *  you can redistribute it and/or modify it under the terms of 
 *  the GNU General Public License as published by the Free Software 
 *  Foundation; either version 2 of the License, or (at your option) 
 *  any later version.
 *
 *  Felimage Noise Plugin for the Gimp is distributed in the hope 
 *  that it will be useful, but WITHOUT ANY WARRANTY; without even 
 *  the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 *  PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with fimg-noise; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
*/

/*
 *  This file has been adapted from the cellular texture basis function 
 *  source by Steven Worley, with the original copyight as follows:
 */
 
/* Copyright 1994, 2002 by Steven Worley
   This software may be modified and redistributed without restriction
   provided this comment header remains intact in the source code.
   This code is provided with no warrantee, express or implied, for
   any purpose.
   
   A detailed description and application examples can be found in the
   1996 SIGGRAPH paper "A Cellular Texture Basis Function" and
   especially in the 2002 book "Texturing and Modeling, a Procedural
   Approach, 3rd edition." There is also extra information on the web
   site http://www.worley.com/cellular.html .

   If you do find interesting uses for this tool, and especially if
   you enhance it, please drop me an email at steve@worley.com. */

#include <math.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>


#ifdef CALIBRATE
#include <glib.h>
#include <string.h>
#else
#include <libgimp/gimp.h>
#endif

#include "poisson.h"
#include "cell_int.h" 
#include "cell.h" 
#include "random.h"

#define DENSITY_ADJUSTMENT  1.0
/*0.398150*/

static void AddSamples_5D(gint32 xi, gint32 yi, gint32 zi, gint32 si, gint32 ti,
		gint32 max_order, double at[5], double *F, 
		double (*delta)[5], guint32 *ID, 
		int order[], CellBasisCache5D *cache );

CellBasisCache5D *InitCellBasis5D() {
	CellBasisCache5D *cache;

	cache = calloc(1,sizeof(CellBasisCache5D) * (3*3*3*3*3)); 

	return cache;
}

void FinishCellBasis5D(CellBasisCache5D *cache) {
	free(cache);
}

void Cells5D(double a0, double a1, double a2, double a3, double a4, gint32 max_order, double *f, double (*p_delta)[5], guint32 *p_id, CellBasisCache5D *cache) {
	double pa0,pa1,pa2,pa3,pa4, ma0,ma1,ma2,ma3,ma4;
	double new_at[5];
	gint32 i,j;
	gint32 int_at[5];
	gint32 int_at_p[5];
	gint32 int_at_m[5];
	double *f_max;
	double delta[4][5];
	guint32 id[4];

	int order[]={0,1,2,3,4};

	for (i=0; i<max_order; i++) f[i]=999999.9;

	new_at[0]=DENSITY_ADJUSTMENT*a0;
	new_at[1]=DENSITY_ADJUSTMENT*a1;
	new_at[2]=DENSITY_ADJUSTMENT*a2;
	new_at[3]=DENSITY_ADJUSTMENT*a3;
	new_at[4]=DENSITY_ADJUSTMENT*a4;

	int_at[0]=(new_at[0]<0.0) ? (gint32)new_at[0]-1 : (gint32)new_at[0]; 
	int_at[1]=(new_at[1]<0.0) ? (gint32)new_at[1]-1 : (gint32)new_at[1]; 
	int_at[2]=(new_at[2]<0.0) ? (gint32)new_at[2]-1 : (gint32)new_at[2]; 
	int_at[3]=(new_at[3]<0.0) ? (gint32)new_at[3]-1 : (gint32)new_at[3]; 
	int_at[4]=(new_at[4]<0.0) ? (gint32)new_at[4]-1 : (gint32)new_at[4]; 

	ma0=new_at[0]-int_at[0];
	ma1=new_at[1]-int_at[1];
	ma2=new_at[2]-int_at[2];
	ma3=new_at[3]-int_at[3];
	ma4=new_at[4]-int_at[4];
	pa0=(1.0-ma0)*(1.0-ma0);
	pa1=(1.0-ma1)*(1.0-ma1);
	pa2=(1.0-ma2)*(1.0-ma2);
	pa3=(1.0-ma3)*(1.0-ma3);
	pa4=(1.0-ma4)*(1.0-ma4);
	ma0*=ma0;
	ma1*=ma1;
	ma2*=ma2;
	ma3*=ma3;
	ma4*=ma4;

	int_at_p[0] = int_at[0]+1;
	int_at_p[1] = int_at[1]+1;
	int_at_p[2] = int_at[2]+1;
	int_at_p[3] = int_at[3]+1;
	int_at_p[4] = int_at[4]+1;

	int_at_m[0] = int_at[0]-1;
	int_at_m[1] = int_at[1]-1;
	int_at_m[2] = int_at[2]-1;
	int_at_m[3] = int_at[3]-1;
	int_at_m[4] = int_at[4]-1;

	f_max = f+(max_order-1);

	/* as generated by gen_tests.py */

	AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache);

	if (pa0 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 1 );
	if (pa1 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 2 );
	if (pa2 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 3 );
	if (pa3 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 4 );
	if (pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 5 );
	if (ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 6 );
	if (ma3 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 7 );
	if (ma2 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 8 );
	if (ma1 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 9 );
	if (ma0 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 10 );
	if (pa0 + pa1 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 11 );
	if (pa0 + pa2 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 12 );
	if (pa0 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 13 );
	if (pa0 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 14 );
	if (pa0 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 15 );
	if (pa0 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 16 );
	if (pa0 + ma2 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 17 );
	if (pa0 + ma1 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 18 );
	if (pa1 + pa2 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 19 );
	if (pa1 + pa3 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 20 );
	if (pa1 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 21 );
	if (pa1 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 22 );
	if (pa1 + ma3 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 23 );
	if (pa1 + ma2 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 24 );
	if (pa2 + pa3 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 25 );
	if (pa2 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 26 );
	if (pa2 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 27 );
	if (pa2 + ma3 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 28 );
	if (pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 29 );
	if (pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 30 );
	if (ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 31 );
	if (ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 32 );
	if (ma2 + pa3 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 33 );
	if (ma2 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 34 );
	if (ma2 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 35 );
	if (ma2 + ma3 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 36 );
	if (ma1 + pa2 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 37 );
	if (ma1 + pa3 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 38 );
	if (ma1 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 39 );
	if (ma1 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 40 );
	if (ma1 + ma3 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 41 );
	if (ma1 + ma2 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 42 );
	if (ma0 + pa1 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 43 );
	if (ma0 + pa2 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 44 );
	if (ma0 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 45 );
	if (ma0 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 46 );
	if (ma0 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 47 );
	if (ma0 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 48 );
	if (ma0 + ma2 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 49 );
	if (ma0 + ma1 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 50 );
	if (pa0 + pa1 + pa2 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 51 );
	if (pa0 + pa1 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 52 );
	if (pa0 + pa1 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 53 );
	if (pa0 + pa1 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 54 );
	if (pa0 + pa1 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 55 );
	if (pa0 + pa1 + ma2 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 56 );
	if (pa0 + pa2 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 57 );
	if (pa0 + pa2 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 58 );
	if (pa0 + pa2 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 59 );
	if (pa0 + pa2 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 60 );
	if (pa0 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 61 );
	if (pa0 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 62 );
	if (pa0 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 63 );
	if (pa0 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 64 );
	if (pa0 + ma2 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 65 );
	if (pa0 + ma2 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 66 );
	if (pa0 + ma2 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 67 );
	if (pa0 + ma2 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 68 );
	if (pa0 + ma1 + pa2 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 69 );
	if (pa0 + ma1 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 70 );
	if (pa0 + ma1 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 71 );
	if (pa0 + ma1 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 72 );
	if (pa0 + ma1 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 73 );
	if (pa0 + ma1 + ma2 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 74 );
	if (pa1 + pa2 + pa3 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 75 );
	if (pa1 + pa2 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 76 );
	if (pa1 + pa2 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 77 );
	if (pa1 + pa2 + ma3 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 78 );
	if (pa1 + pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 79 );
	if (pa1 + pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 80 );
	if (pa1 + ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 81 );
	if (pa1 + ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 82 );
	if (pa1 + ma2 + pa3 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 83 );
	if (pa1 + ma2 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 84 );
	if (pa1 + ma2 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 85 );
	if (pa1 + ma2 + ma3 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 86 );
	if (pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 87 );
	if (pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 88 );
	if (pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 89 );
	if (pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 90 );
	if (ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 91 );
	if (ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 92 );
	if (ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 93 );
	if (ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 94 );
	if (ma1 + pa2 + pa3 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 95 );
	if (ma1 + pa2 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 96 );
	if (ma1 + pa2 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 97 );
	if (ma1 + pa2 + ma3 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 98 );
	if (ma1 + pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 99 );
	if (ma1 + pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 100 );
	if (ma1 + ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 101 );
	if (ma1 + ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 102 );
	if (ma1 + ma2 + pa3 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 103 );
	if (ma1 + ma2 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 104 );
	if (ma1 + ma2 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 105 );
	if (ma1 + ma2 + ma3 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 106 );
	if (ma0 + pa1 + pa2 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 107 );
	if (ma0 + pa1 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 108 );
	if (ma0 + pa1 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 109 );
	if (ma0 + pa1 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 110 );
	if (ma0 + pa1 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 111 );
	if (ma0 + pa1 + ma2 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 112 );
	if (ma0 + pa2 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 113 );
	if (ma0 + pa2 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 114 );
	if (ma0 + pa2 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 115 );
	if (ma0 + pa2 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 116 );
	if (ma0 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 117 );
	if (ma0 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 118 );
	if (ma0 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 119 );
	if (ma0 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 120 );
	if (ma0 + ma2 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 121 );
	if (ma0 + ma2 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 122 );
	if (ma0 + ma2 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 123 );
	if (ma0 + ma2 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 124 );
	if (ma0 + ma1 + pa2 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 125 );
	if (ma0 + ma1 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 126 );
	if (ma0 + ma1 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 127 );
	if (ma0 + ma1 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 128 );
	if (ma0 + ma1 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 129 );
	if (ma0 + ma1 + ma2 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 130 );
	if (pa0 + pa1 + pa2 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 131 );
	if (pa0 + pa1 + pa2 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 132 );
	if (pa0 + pa1 + pa2 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 133 );
	if (pa0 + pa1 + pa2 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 134 );
	if (pa0 + pa1 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 135 );
	if (pa0 + pa1 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 136 );
	if (pa0 + pa1 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 137 );
	if (pa0 + pa1 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 138 );
	if (pa0 + pa1 + ma2 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 139 );
	if (pa0 + pa1 + ma2 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 140 );
	if (pa0 + pa1 + ma2 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 141 );
	if (pa0 + pa1 + ma2 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 142 );
	if (pa0 + pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 143 );
	if (pa0 + pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 144 );
	if (pa0 + pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 145 );
	if (pa0 + pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 146 );
	if (pa0 + ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 147 );
	if (pa0 + ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 148 );
	if (pa0 + ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 149 );
	if (pa0 + ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 150 );
	if (pa0 + ma1 + pa2 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 151 );
	if (pa0 + ma1 + pa2 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 152 );
	if (pa0 + ma1 + pa2 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 153 );
	if (pa0 + ma1 + pa2 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 154 );
	if (pa0 + ma1 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 155 );
	if (pa0 + ma1 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 156 );
	if (pa0 + ma1 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 157 );
	if (pa0 + ma1 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 158 );
	if (pa0 + ma1 + ma2 + pa3 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 159 );
	if (pa0 + ma1 + ma2 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 160 );
	if (pa0 + ma1 + ma2 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 161 );
	if (pa0 + ma1 + ma2 + ma3 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 162 );
	if (pa1 + pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 163 );
	if (pa1 + pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 164 );
	if (pa1 + pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 165 );
	if (pa1 + pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 166 );
	if (pa1 + ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 167 );
	if (pa1 + ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 168 );
	if (pa1 + ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 169 );
	if (pa1 + ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 170 );
	if (ma1 + pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 171 );
	if (ma1 + pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 172 );
	if (ma1 + pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 173 );
	if (ma1 + pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 174 );
	if (ma1 + ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 175 );
	if (ma1 + ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 176 );
	if (ma1 + ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 177 );
	if (ma1 + ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 178 );
	if (ma0 + pa1 + pa2 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 179 );
	if (ma0 + pa1 + pa2 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 180 );
	if (ma0 + pa1 + pa2 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 181 );
	if (ma0 + pa1 + pa2 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 182 );
	if (ma0 + pa1 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 183 );
	if (ma0 + pa1 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 184 );
	if (ma0 + pa1 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 185 );
	if (ma0 + pa1 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 186 );
	if (ma0 + pa1 + ma2 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 187 );
	if (ma0 + pa1 + ma2 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 188 );
	if (ma0 + pa1 + ma2 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 189 );
	if (ma0 + pa1 + ma2 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 190 );
	if (ma0 + pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 191 );
	if (ma0 + pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 192 );
	if (ma0 + pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 193 );
	if (ma0 + pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 194 );
	if (ma0 + ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 195 );
	if (ma0 + ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 196 );
	if (ma0 + ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 197 );
	if (ma0 + ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 198 );
	if (ma0 + ma1 + pa2 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 199 );
	if (ma0 + ma1 + pa2 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 200 );
	if (ma0 + ma1 + pa2 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 201 );
	if (ma0 + ma1 + pa2 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 202 );
	if (ma0 + ma1 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 203 );
	if (ma0 + ma1 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 204 );
	if (ma0 + ma1 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 205 );
	if (ma0 + ma1 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 206 );
	if (ma0 + ma1 + ma2 + pa3 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 207 );
	if (ma0 + ma1 + ma2 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 208 );
	if (ma0 + ma1 + ma2 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 209 );
	if (ma0 + ma1 + ma2 + ma3 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at[4], max_order, new_at, f, delta, id, order, cache + 210 );
	if (pa0 + pa1 + pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 211 );
	if (pa0 + pa1 + pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 212 );
	if (pa0 + pa1 + pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 213 );
	if (pa0 + pa1 + pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 214 );
	if (pa0 + pa1 + ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 215 );
	if (pa0 + pa1 + ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 216 );
	if (pa0 + pa1 + ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 217 );
	if (pa0 + pa1 + ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 218 );
	if (pa0 + ma1 + pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 219 );
	if (pa0 + ma1 + pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 220 );
	if (pa0 + ma1 + pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 221 );
	if (pa0 + ma1 + pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 222 );
	if (pa0 + ma1 + ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 223 );
	if (pa0 + ma1 + ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 224 );
	if (pa0 + ma1 + ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 225 );
	if (pa0 + ma1 + ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_p[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 226 );
	if (ma0 + pa1 + pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 227 );
	if (ma0 + pa1 + pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 228 );
	if (ma0 + pa1 + pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 229 );
	if (ma0 + pa1 + pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 230 );
	if (ma0 + pa1 + ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 231 );
	if (ma0 + pa1 + ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 232 );
	if (ma0 + pa1 + ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 233 );
	if (ma0 + pa1 + ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_p[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 234 );
	if (ma0 + ma1 + pa2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 235 );
	if (ma0 + ma1 + pa2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 236 );
	if (ma0 + ma1 + pa2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 237 );
	if (ma0 + ma1 + pa2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_p[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 238 );
	if (ma0 + ma1 + ma2 + pa3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 239 );
	if (ma0 + ma1 + ma2 + pa3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at_p[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 240 );
	if (ma0 + ma1 + ma2 + ma3 + pa4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at_p[4], max_order, new_at, f, delta, id, order, cache + 241 );
	if (ma0 + ma1 + ma2 + ma3 + ma4 < *f_max) AddSamples_5D( int_at_m[0], int_at_m[1], int_at_m[2], int_at_m[3], int_at_m[4], max_order, new_at, f, delta, id, order, cache + 242 );

	for (i=0; i<max_order; i++) {
		f[i]=sqrt(f[i])*(1.0/DENSITY_ADJUSTMENT);      
		j = order[i];
		p_delta[i][0]=delta[j][0]*(1.0/DENSITY_ADJUSTMENT);
		p_delta[i][1]=delta[j][1]*(1.0/DENSITY_ADJUSTMENT);
		p_delta[i][2]=delta[j][2]*(1.0/DENSITY_ADJUSTMENT);
		p_delta[i][3]=delta[j][3]*(1.0/DENSITY_ADJUSTMENT);
		p_delta[i][4]=delta[j][4]*(1.0/DENSITY_ADJUSTMENT);
		p_id[i] = id[j];
	}

	return;
}


static void AddSamples_5D(gint32 xi, gint32 yi, gint32 zi, gint32 si, gint32 ti,
		gint32 max_order, double at[5], double *F, 
		double (*delta)[5], guint32 *ID, 
		int order[], CellBasisCache5D *cache ) {

	double dx, dy, dz, ds, dt, fx, fy, fz, fs, ft, d2;
	gint32 count, i, j, index;
	int slot;
	guint32 seed, this_id;

	seed=Hash5(xi,yi,zi,si,ti);
	count=Poisson_count[seed&255];

	if ( (seed != cache->last_seed) || (seed == 0)) {
		cache->last_seed = seed;	

		seed=RANDOM(seed);
		for (j=0; j<count; j++) {
			cache->id[j] = this_id = seed;
			
			seed=RANDOM(seed);
			fx = (seed+0.5)*(1.0/4294967296.0); 
			
			seed=RANDOM(seed);
			fy = (seed+0.5)*(1.0/4294967296.0);
			
			seed=RANDOM(seed);
			fz =(seed+0.5)*(1.0/4294967296.0);

			seed=RANDOM(seed);
			fs =(seed+0.5)*(1.0/4294967296.0);
			
			seed=RANDOM(seed);
			ft =(seed+0.5)*(1.0/4294967296.0);
			
			seed=RANDOM(seed);

			dx = (cache->p[0][j] = fx)+xi-at[0]; 
			dy = (cache->p[1][j] = fy)+yi-at[1];
			dz = (cache->p[2][j] = fz)+zi-at[2];
			ds = (cache->p[3][j] = fs)+si-at[3];
			dt = (cache->p[4][j] = ft)+ti-at[4];

			d2 = dx*dx+dy*dy+dz*dz+ds*ds+dt*dt;

			i = max_order-1;
			
			if (d2<F[i]) {
				
				index=max_order;
				while (index>0 && d2<F[index-1]) index--;

				slot = order[i];

				while ((i--)>index) {
					F[i+1]=F[i];
					order[i+1] = order[i];
				}		
				
				F[index]=d2;
				order[index] = slot;
					
				ID   [slot]=this_id;
				delta[slot][0]=dx;
				delta[slot][1]=dy;
				delta[slot][2]=dz;
				delta[slot][3]=ds;
				delta[slot][4]=dt;
			}
		}
	} else {
		for (j=0; j<count; j++) {
			dx = cache->p[0][j]+xi-at[0];
			dy = cache->p[1][j]+yi-at[1];
			dz = cache->p[2][j]+zi-at[2];
			ds = cache->p[3][j]+si-at[3];
			dt = cache->p[4][j]+ti-at[4];
			
			d2 = dx*dx+dy*dy+dz*dz+ds*ds+dt*dt;

			i = max_order-1;
			
			if (d2<F[i]) {
				index=max_order;
				while (index>0 && d2<F[index-1]) index--;

				slot = order[i];

				while ((i--)>index) {
					F[i+1]=F[i];
					order[i+1] = order[i];
				}		
				
				F[index]=d2;
				order[index] = slot;
					
				ID   [slot]= cache->id[j];
				delta[slot][0]=dx;
				delta[slot][1]=dy;
				delta[slot][2]=dz;
				delta[slot][3]=ds;
				delta[slot][4]=dt;
			}
		}
	}

	return;
}

